from __future__ import annotations

import networkx as nx

from pop2net.location import Location
import pop2net.utils as utils

from . import actor as _actor


class LocationDesigner(Location):
    """Helper class to create locations from inside the Creator."""

    label: str | None = None
    location_class: type | None = None
    location_name: str | None = None

    n_actors: int | None = None
    overcrowding: bool = None
    only_exact_n_actors: bool = False
    n_locations: int | None = None
    static_weight: bool = False
    recycle: bool = True
    nxgraph: nx.Graph | None = None

    def __init__(self, *args, **kwargs) -> None:
        """Create a helper class to create locations."""
        super().__init__(*args, **kwargs)
        self.group_id: int | None = None
        self.subgroup_id: int | None = None
        self.group_value: int | str | None = None
        self.subgroup_value: int | str | None = None

    def setup(self) -> None:
        """Use this method to set instance attributes, for instance.

        This method is called automatically by the creator after creating an instance.
        """

    def filter(self, actor: _actor.Actor) -> bool:  # noqa: ARG002
        """Check whether the actor is meant to join this type of location.

        Args:
            actor: The actor that is currently processed by the Creator.

        Returns:
            True if the actor is allowed to join the location, False otherwise.
            Returns True by default.
        """
        return True

    def bridge(self, actor: _actor.Actor) -> float | str | list | None:  # noqa: ARG002
        """Create locations with one actor for each unique value returned.

        Cannot be used in combination with melt().

        Args:
            actor (_actor.Actor): The actor that is currently processed by the Creator.

        Returns:
            float | str | list | None:
                The value which is used to assign actors to location instances.
        """
        return None

    def split(self, actor: _actor.Actor) -> float | str | list | None:  # noqa: ARG002
        """Creates seperate location instances for each unique returned value.

        Args:
            actor: The actor that is currently processed by the Creator.

        Returns:
            float | str | list | None:
                The value(s) that determine(s) to which location instance the actor is assigned.
        """
        return None

    def stick_together(self, actor: _actor.Actor) -> float | str:
        """Assigns actors with a shared value on an attribute to the same location instance.

        Args:
            actor (_actor.Actor): The actor that is currently processed by the Creator.

        Returns:
            float | str: A value that defines the groups of actors.
        """
        return actor.id_p2n

    def nest(self) -> type[Location] | None:
        """Nests this location class into another location class.

        Ensures that the actors assigned to the same instance of this location class
        are also assigned to the same instance of the returned location class.

        Returns:
           type[Location] | None: The location class in which this location class should be nested.
        """
        return None

    def melt(self) -> list[Location] | tuple[Location]:
        """Allows to assign actors with different attribute values to the same location.

        Merges the actors assigned to the instances of the returned location classes
        into one instance of this location class.

        Returns:
            None | list | tuple: A list or tuple of location classes.
        """
        return []

    def _subsplit(self, actor: _actor.Actor) -> str | float | list | None:  # noqa: ARG002
        """Splits a location instance into sub-instances to create a certain network structure.

        Args:
            actor (_actor.Actor): The actor that is currently processed by the Creator.

        Returns:
            str | float | None: A value or a list of values that represent specific sub-instances.
        """
        if self.nxgraph is None:
            return None
        else:
            node_indices = list(self.nxgraph.nodes)
            actor_pos = self.actors_.index(actor)

            if actor_pos <= len(node_indices) - 1:
                node_index = node_indices[actor_pos]

                return [
                    utils._join_positions(pos1=node_index, pos2=neighbor)
                    for neighbor in self.nxgraph.neighbors(node_index)
                ]

            else:
                return None

    def mutate(self) -> None | dict[str:list]:
        """Creates new versions of this location designer with different attributes.

        Returns:
            dict: A dictionary that specifies the values for each mutation.
        """
        return None


class MeltLocationDesigner(Location):
    """Helper class to melt locations."""

    label: str | None = None

    location_class: type | None = None
    location_name: str | None = None
    n_actors: int | None = None
    overcrowding: bool | None = None
    only_exact_n_actors: bool = False
    n_locations: int | None = None

    def filter(self, actor: _actor.Actor) -> bool:  # noqa: ARG002
        """Check whether the actor is meant to join this type of location.

        This is a boilerplate implementation of this method which always returns True; i.e. all
        actors will always be allowed at this location. Override this method in your own
        implementations as you seem fit.

        Args:
            actor: The actor that is currently processed by the Creator.

        Returns:
            True if the actor is allowed to join the location, False otherwise.
        """
        return True

    def stick_together(self, actor: _actor.Actor) -> float | str:
        """Assigns actors with a shared value on an attribute to the same location instance.

        Args:
            actor (_actor.Actor): The actor that is currently processed by the Creator.

        Returns:
            float | str: A value that defines the groups of actors.
        """
        return actor.id_p2n

    def split(self, actor: _actor.Actor) -> float | str | list | None:  # noqa: ARG002
        """Creates seperate location instances for each unique returned value.

        Args:
            actor: The actor that is currently processed by the Creator.

        Returns:
            float | str | list | None: The value(s) that determine(s) to which location instance
                the actor is assigned.
        """
        return None

    def weight(self, actor: _actor.Actor) -> float | None:  # noqa: ARG002
        """Defines the edge weight between the actor and the location instance.

        Defines how the edge weight between an actor and the location is determined.
        This is a boilerplate implementation of this method which always returns 1; i.e. all
        edge weights will be 1. Override this method in your own implementations as you seem fit.

        Args:
            actor: The actor that is currently processed by the Creator.

        Returns:
            The edge weight.
        """
        return None



"""Base class to create Actor objects."""

from __future__ import annotations

import typing
import warnings

if typing.TYPE_CHECKING:
    from . import location as _location


class Actor:
    """This is a Base class to represent actors in the simulation.

    Actors' behavior can be implemented in classes that inherit from this.
    """

    def __init__(self, *args, **kwargs) -> None:
        """Actor Constructor."""
        self.env = None
        self.id_p2n = None
        self.model = None
        self.type = type(self).__name__

        super().__init__(*args, **kwargs)

    def neighbors(self, location_labels: list[str] | None = None) -> list:
        """Return all neighbors of an actor.

        Convenience method that returns all neighbors over all locations this actor is currently
        located in. The locations to be considered can be defined with location_labels.

        Args:
            location_labels: A list of location_labels.

        Returns:
            All actors co-located with this actor over all locations.
        """
        return self.env.neighbors_of_actor(self, location_labels=location_labels)

    def shared_locations(self, actor, location_labels: list[str] | None = None) -> list:
        """Returns all locations that this actor shares with another actor.
        Use location_labels to specify which type of locations should be included in the search
        for shared locations.

        Args:
            actor (p2n.Actor): The other actor.
            location_labels (list[str] | None, optional): A list of location_labels. Defaults to None.

        Returns:
            list: Shared locations.
        """
        return self.env.locations_between_actors(
            actor1=self,
            actor2=actor,
            location_labels=location_labels,
        )

    def add_location(self, location: _location.Location, weight: float | None = None) -> None:
        """Add this actor to a given location.
        If weight is None, it will be set to 1.

        Args:
            location: Add actor to this location.
            weight (float | None): The edge weight between the actor and the location.
                Defaults to None.
        """
        self.env.add_actor_to_location(actor=self, location=location, weight=weight)

    def add_locations(self, locations: list, weight: float | None = None) -> None:
        """Add this actor to multiple locations.

        Args:
            locations (list): Add the actor to these locations.
            weight (float | None): The edge weight between the actor and the location.
                Defaults to None.
        """
        for location in locations:
            self.add_location(location=location, weight=weight)

    def remove_location(self, location: _location.Location) -> None:
        """Remove this actor from a given location.

        Args:
            location: Remove actor from this location.
        """
        self.env.remove_actor_from_location(self, location)

    def remove_locations(self, locations: list) -> None:
        """Remove this Actor from the given locations.

        Args:
            locations (list): A list of location instances.
        """
        for location in locations:
            self.remove_location(location)

    @property
    def locations(self) -> list:
        """Returns a list of locations that this actor is associated with.

        Returns:
            A list of locations.
        """
        return self.env.locations_of_actor(self)

    @property
    def location_labels(self) -> list[str]:
        """Returns a list of labels of the locations this actor is associated with.

        Returns:
            list[str]: A list of location labels.
        """
        return [location.label for location in self.locations]

    def get_actor_weight(self, actor: Actor, location_labels: list | None = None) -> float:
        """Returns the edge weight between this actor and a given other actor.

        This is summed over all shared locations.

        Args:
            actor: The other actor.
            location_labels (list): A list of location classes to specify the type of locations
                which are considered.

        Returns:
            A weight of the contact between the two actors.
        """
        weight = 0
        for location in self.shared_locations(actor=actor, location_labels=location_labels):
            weight += location.project_weights(actor1=self, actor2=actor)
        return weight

    def get_location_weight(self, location) -> float:
        """Return the edge weight between this actor and a given location.

        Args:
            location (_type_): A location.

        Returns:
            float: The edge weight.
        """
        return self.env.get_weight(actor=self, location=location)

    def connect(self, actor: Actor, location_cls: _location = None, weight: float | None = None):
        """Connects this actor with a given other actor via an instance of a given location class.
        If location_cls is None, the default pop2net.Location class is used to create a new location
        instance. If weight is None, it will be set to 1.

        Args:
            actor (p2n.Actor): An actor to connect with.
            location_cls (type): The location class that is used to create a location instance.
            weight(float | None): The edge weight between the actors and the location.
                Defaults to None.
        """
        self.env.connect_actors(actors=[self, actor], location_cls=location_cls, weight=weight)

    def disconnect(
        self,
        neighbor: Actor,
        location_labels: list | None = None,
        remove_self=True,
        remove_neighbor=True,
        remove_locations: bool = False,
    ):
        """Disconnects this actor from a given other actor by removing them from shared locations.

        If a list of location types is given, only shared locations of the given types are
        considered. Turn on `remove_locations` in order to not only remove the actors from the
        given location instance but also to remove the location instance from the model.  Keep in
        mind that this may affect other actors that are still connected with the location instance.

        Args:
            neighbor (Actor): An actor to disconnect from.
            location_labels (list | None, optional): A list of location types to specify which
            shared locations are considered. Defaults to None.
            remove_self (bool): Should the actor be removed from the shared locations?
                Defaults to True.
            remove_neighbor (bool): Should the neighbor be removed from the shared locations?
                Defaults to True.
            remove_locations (bool, optional): A bool that determines whether the shared locations
                shall be removed from the model. Defaults to False.
        """
        shared_locations = self.shared_locations(
            actor=neighbor,
            location_labels=location_labels,
        )

        for location in shared_locations:
            warn = False
            for actor in location.actors:
                if actor not in [self, neighbor]:
                    warn = True
                    break

            if remove_self:
                location.remove_actor(self)

                if warn:
                    msg = (
                        "There are other actors at the location from which you have removed actors."
                    )
                    warnings.warn(msg)

            if remove_neighbor:
                location.remove_actor(neighbor)

                if warn:
                    msg = (
                        "There are other actors at the location from which you have removed actors."
                    )
                    warnings.warn(msg)

            if remove_locations:
                self.env.remove_location(location)

                if warn:
                    msg = "You have removed a location to which other actors were still connected."
                    warnings.warn(msg)



"""Pop2net. An extension package for location-based simulations based on AgentPy."""

from .actor import Actor
from .creator import Creator
from .entity_list import EntityList
from .environment import Environment
from .exceptions import Pop2netException
from .inspector import NetworkInspector
from .location import Location
from .location_designer import LocationDesigner
from .location_designer import MeltLocationDesigner

# from .sequences import LocationList

__all__ = [
    "Actor",
    "EntityList",
    "Pop2netException",
    "Location",
    "LocationDesigner",
    "MeltLocationDesigner",
    "NetworkInspector",
    "Environment",
    "LocationList",
    "Creator",
]



"""Helper class to provide fake SOEP data based on the fakers package."""

from typing import Any
from typing import Literal

from faker.providers import BaseProvider
import numpy as np

from . import constants as c


class SOEPProvider(BaseProvider):
    """Create Fake Data that is as close to the SOEP as possible.

    Parameters of the random distributions are taken from the original dataset.
    """

    def __init__(self, generator: Any) -> None:
        """Instatiate the SOEP faker class.

        Args:
            generator (Any): Can be used to set the seed.
        """
        # TODO: loose type hint on generator...
        self.generator = generator
        self.rng = np.random.default_rng(self.generator._global_seed)

        # For hhid
        self.hh_type_dist = lambda: self._multinomial_select(c.HH_TYPE)
        self.hh_size_dist = lambda: self.rng.poisson(lam=c.HH_SIZE["lam"])
        self.hh_n_children_dist = lambda: self._multinomial_select(c.HH_N_CHILDREN)
        self.hh_cur = 0
        self.used_hhids: set[int] = set()

        # For age
        self.age_child_dist = lambda: self.rng.normal(
            loc=c.AGE_DIST["child"]["loc"],
            scale=c.AGE_DIST["child"]["loc"],
        )
        self.age_adult_dist = lambda: self.rng.normal(
            loc=c.AGE_DIST["adult"]["loc"],
            scale=c.AGE_DIST["adult"]["loc"],
        )
        self.age_senior_dist = lambda: self.rng.normal(
            loc=c.AGE_DIST["senior"]["loc"],
            scale=c.AGE_DIST["senior"]["loc"],
        )

        # For gender
        self.gender_dist = lambda: self._multinomial_select(c.GENDER_DIST)

        # For work hours per day
        self.work_at_all_dist = lambda: self.rng.binomial(n=1, p=c.WORKING_HOURS["at_all"])
        self.work_full_day_dist = lambda: self.rng.binomial(n=1, p=c.WORKING_HOURS["8_hours"])
        self.work_dist = lambda: self.rng.normal(
            loc=c.WORKING_HOURS["other"]["loc"],  # type: ignore
            scale=c.WORKING_HOURS["other"]["scale"],  # type: ignore
        )

        # type of work
        self.nace2_dist = lambda: self._multinomial_select(c.NACE2_DIVISIONS)

    def _multinomial_select(self, dist: dict):
        pvals = list(dist.values())
        labels = dict(enumerate(dist.keys()))

        index = int(np.argmax(self.rng.multinomial(n=1, pvals=pvals)))

        return labels[index]

    def _hhid(self) -> int:
        counter = 0
        while True:
            hhid = self.generator.random_int()
            if hhid not in self.used_hhids:
                return hhid
            counter += 1
            if counter > 1000:
                msg = "Could not find new hhid after 1000 iterations!"
                raise ValueError(msg)

    def household(self) -> tuple[int, str, int, int]:
        """Create correlated fake numbers on household level.

        Returns:
            Tuple[int, str, int, int]: [hh_id, hh_type, hh_size, n_children]
        """
        hh_type = self.hh_type_dist()

        if hh_type == "children":
            min_hh_size = 2
        else:
            min_hh_size = 1

        while (hh_size := self.hh_size_dist()) < min_hh_size:
            pass

        if hh_type != "children":
            n_children = 0
        else:
            while (n_children := self.hh_n_children_dist()) >= hh_size:
                pass

        return (self._hhid(), hh_type, hh_size, n_children)

    def person(self, person: Literal["child", "adult", "senior"]):
        """Create a simulated single persion.

        Args:
            person (Literal["child", "adult", "senior"]): Type of person.

        Returns:
            dict: Attributes of the simulated person.
        """
        work_hours = self.work_hours_per_day(person=person)
        nace2 = self.nace2_division(work_hours)
        case = {
            "age": self.age(person=person),
            "gender": self.gender(),
            "work_hours_day": work_hours,
            "nace2_division": nace2,
        }
        return case

    def household_persons(self) -> list[dict]:
        """Create a household full of simulated persons.

        Returns:
            List[Dict]: List of dicts with attributes of the simulated persons.
        """
        hh_id, hh_type, hh_size, n_children = self.household()

        if hh_type == "no_children":
            persons = [self.person("adult") for _ in range(hh_size)]
        elif hh_type == "senior":
            persons = [self.person("senior") for _ in range(hh_size)]
        else:
            n_adults = hh_size - n_children
            persons = [self.person("adult") for _ in range(n_adults)]
            persons.extend([self.person("child") for _ in range(n_children)])

        for person in persons:
            person["hid"] = hh_id

        return persons

    def age(self, person: Literal["child", "adult", "senior"] = "adult") -> int:
        """Draw the age for a specific person type.

        Args:
            person (Literal["child", "adult", "senior"], optional): Person type that is the basis
            for the draw. Defaults to "adult".

        Returns:
            int: The age.
        """
        if person == "child":
            min_age = 0.0
            max_age = 18
            dist = self.age_child_dist
        elif person == "adult":
            min_age = 18
            max_age = 100
            dist = self.age_adult_dist
        elif person == "senior":
            min_age = 40
            max_age = 100
            dist = self.age_senior_dist

        # iterate until a positive number
        while not min_age <= (age := dist()) <= max_age:
            pass

        return round(age, 0)

    def work_hours_per_day(self, person: Literal["child", "adult", "senior"]) -> float:
        """Draw the amount of work hours per day based on the type of person.

        The underlying distributions are roughly equal to the ones discovered in SOEP.

        Args:
            person (Literal["child", "adult", "senior"]): Type of person.

        Returns:
            float: Number of work hours per day.
        """
        # children do not work
        if person != "adult":
            return 0.0

        # about half of people work at all
        if self.work_at_all_dist() == 0:
            return 0.0

        # of the remaining, about 19% have a full work day (8 hours)
        if self.work_full_day_dist() == 1:
            return 8.0

        # rest is drawn for normal (more like bimodal in SOEP but who cares..)
        while not 0.0 < (hours := self.work_dist()) < 16.0:
            pass

        return round(hours, 6)

    def gender(self) -> str:
        """Draw a gender based on the gender distribution from SOEP.

        Returns:
            str: The drawn gender. Possible values are: ["female", "male", "other"].
        """
        return self.gender_dist()

    def nace2_division(self, work_hours: float) -> int:
        """Draw a NACE2 division based on SOEP.

        If the work hours are 0, the NACE2 code is always -2.

        Args:
            work_hours (float): Work hours. If 0 the return value is always -2.

        Returns:
            int: NACE2 division.
        """
        if work_hours > 0:
            return self.nace2_dist()
        else:
            return -2



"""The environment class. It encapsulates the full simulation."""

from __future__ import annotations

import itertools
import typing
import warnings

import networkx as nx

if typing.TYPE_CHECKING:
    from . import actor as _actor
    from . import location as _location

import pop2net as p2n


class Environment:
    """Class that organizes actors and locations."""

    def __init__(self, model=None, framework: str | None = None, enable_p2n_warnings=True):
        """Initialize a new environment.

        Args:
            model (agentpy.Model | mesa.Model, optional): A simulation model. Defaults to None.
            framework (str | None, optional): The ABM-framework you you want to use.
                Options are: "agentpy" & "mesa". Defaults to None.
            enable_p2n_warnings (bool, optional): Should pop2net warnings be shown? Defaults to True.

        Raises:
            ValueError: _description_
        """
        # settings
        self.framework: str | None = framework
        self._framework = None
        self.enable_p2n_warnings = enable_p2n_warnings

        # import framework dependencies, if required
        if self.framework is None:
            pass
        elif self.framework == "agentpy":
            import agentpy

            self._framework = agentpy
        elif self.framework == "mesa":
            import mesa

            self._framework = mesa
        else:
            raise ValueError(f"Invalid framework selected: {self.framework}")

        # connected objects
        self.model = model
        self.g = nx.Graph()

        # TODO: should we add creator and inspector as attributes?
        # self.creator = Creator(env=self)
        # self.inspector = NetworkInspector(env=self)

        # a unique id that is added to the objects of this environment
        self._fresh_id = 0

    def _attach_fresh_id(self, obj):
        obj.id_p2n = self._fresh_id
        self._fresh_id += 1

    def _to_framework(self, objects):
        if self.framework is None:
            return p2n.EntityList(objects if objects is not None else [])
        elif self.framework == "agentpy":
            return self._framework.AgentList(model=self.model, objs=objects)
        elif self.framework == "mesa":
            return self._framework.agent.AgentSet(agents=objects, random=self.model.random)
        else:
            raise ValueError("Invalid framework.")

    @property
    def actors(self) -> list:
        """Show a iterable view of all actors in the environment.

        Important: While you can make changes to the objects in this list, you
        cannot modify this attribute itself. Instead you have to use methods like
        Environment.add_actor() or Environment.remove_actor(), for instance.

        Returns:
            list: A non-mutable list of all actors in the environment.
        """
        return self._to_framework(
            [data["_obj"] for _, data in self.g.nodes(data=True) if data["bipartite"] == 0]
        )

    @property
    def locations(self) -> list:
        """Show a iterable view of all locations in the environment.

        Important: While you can make changes to the objects in this list, you
        cannot modify this attribute itself. Instead you have to use methods like
        Environment.add_location() or Environment.remove_location(), for instance.

        Returns:
            LocationList: a non-mutable LocationList of all locations in the environment.
        """
        return self._to_framework(
            [data["_obj"] for _, data in self.g.nodes(data=True) if data["bipartite"] == 1]
        )

    # TODO: def add_obj as a common parent method for add_actor & add_location
    def add_actor(self, actor: _actor.Actor) -> None:
        """Add an actor to the environment.

        The added actor will have no connections to other actors or locatons by default.
        If the actor is already in the current environment, this methods does nothing.

        Args:
            actor: Actor to be added to the environment.
        """
        if actor.id_p2n is None:
            self._attach_fresh_id(actor)

        if not self.g.has_node(actor.id_p2n):
            self.g.add_node(actor.id_p2n, bipartite=0, _obj=actor)
            actor.env = self
        else:
            raise ValueError("This environment already has an entity with this id.")

    def add_actors(self, actors: list) -> None:
        """Add actors to the environment.

        Args:
            actors (list): A list of the actors to be added.
        """
        for actor in actors:
            self.add_actor(actor)

    def add_location(self, location: _location.Location) -> None:
        """Add a location to the environment.

        The added location will have no connections to other actors or locatons by default.
        If the location is already in the current environment, this methods does nothing.

        Args:
            location: Location to be added to the environment.
        """
        if location.id_p2n is None:
            self._attach_fresh_id(location)

        if not self.g.has_node(location.id_p2n):
            self.g.add_node(location.id_p2n, bipartite=1, _obj=location)
            location.env = self
        else:
            raise ValueError("This environment already has an entity with this id.")

    def add_locations(self, locations: list) -> None:
        """Add multiple locations to the environment at once.

        Args:
            locations (list): An iterable over multiple locations.
        """
        for location in locations:
            self.add_location(location)

    def add_actor_to_location(
        self,
        location: _location.Location,
        actor: _actor.Actor,
        weight: float | None = None,
        **kwargs,
    ) -> None:
        """Add an actor to a specific location.

        Both the actor and the location have to be defined beforehand. All additional keyword
        arguments will be edge attributes for this connection.

        Args:
            location: Location the actor is to be added to.
            actor: Actor to be added to the location.
            weight: An optional weight for the connection.
            **kwargs: Additional edge attributes.

        Raises:
            Exception: Raised if the location does not exist in the environment.
            Exception: Raised if the actor does not exist in the environment.
        """
        # TODO: Create custom exceptions
        if not self.g.has_node(location.id_p2n):
            msg = f"Location {location} does not exist in Environment!"
            raise Exception(msg)
        if not self.g.has_node(actor.id_p2n):
            msg = f"Actor {actor} does not exist in Environment!"
            raise Exception(msg)

        self.g.add_edge(actor.id_p2n, location.id_p2n, **kwargs)
        self.set_weight(actor=actor, location=location, weight=weight)

    def remove_actor(self, actor: _actor.Actor) -> None:
        """Remove an actor from the environment.

        If the actor does not exist in the environment, this method does nothing.

        Args:
            actor: Actor to be removed.
        """
        if self.g.has_node(actor.id_p2n):
            self.g.remove_node(actor.id_p2n)

    def remove_actors(self, actors: list[_actor.Actor]) -> None:
        """Remove multiple actors from the environment at once.

        Args:
            actors (list): An iterable over multiple actors.
        """
        for actor in actors:
            self.remove_actor(actor)

    def remove_location(self, location: _location.Location) -> None:
        """Remove a location from the environment.

        If the location does not exist in the environment, this method does nothing.

        Args:
            location: Location to be removed.
        """
        if self.g.has_node(location.id_p2n):
            self.g.remove_node(location.id_p2n)

    def remove_locations(self, locations: list[_location.Location]) -> None:
        """Remove multiple locations at once.

        Args:
            locations (list): An iterable over locations.
        """
        for location in locations:
            self.remove_location(location)

    def remove_actor_from_location(
        self,
        location: _location.Location,
        actor: _actor.Actor,
    ) -> None:
        """Remove an actor from a location.

        Args:
            location: Location, the actor is to be removed from.
            actor: Actor to be disassociated with the location.

        Raises:
            Exception: Raised if the location does not exist in the environment.
            Exception: Raised if the actor does not exist in the environment.
        """
        # TODO: use custom exceptions
        if not self.g.has_node(location.id_p2n):
            msg = f"Location {location} does not exist in Environment!"
            raise Exception(msg)
        if not self.g.has_node(actor.id_p2n):
            msg = f"Actor {actor} does not exist in Environment!"
            raise Exception(msg)

        if self.g.has_edge(actor.id_p2n, location.id_p2n):
            self.g.remove_edge(actor.id_p2n, location.id_p2n)

    def actors_of_location(self, location: _location.Location):
        """Return the list of actors associated with a specific location.

        Args:
            location: The desired location.

        Returns:
            A list of actors.
        """
        nodes = self.g.neighbors(location.id_p2n)
        actors = [
            self.g.nodes[node]["_obj"] for node in nodes if self.g.nodes[node]["bipartite"] == 0
        ]
        return self._to_framework(actors)

    def locations_of_actor(self, actor: _actor.Actor) -> list[_location.Location]:
        """Return the list of locations associated with a specific actor.

        Args:
            actor: The desired actor.

        Returns:
            A list of locations.
        """
        nodes = self.g.neighbors(actor.id_p2n)
        locations = [
            self.g.nodes[node]["_obj"] for node in nodes if self.g.nodes[node]["bipartite"] == 1
        ]
        return self._to_framework(locations)

    def neighbors_of_actor(
        self,
        actor: _actor.Actor,
        location_labels: list | None = None,
    ) -> list[_actor.Actor]:
        """Return a list of neighboring actors for a specific actor.

        The locations to be considered can be defined with location_labels.

        Args:
            actor: Actor of whom the neighbors are to be returned.
            location_labels: A list of location_labels.

        Returns:
            The list of neighbors for the specified actor.
        """
        if location_labels:
            locations = (
                node
                for node in self.g.neighbors(actor.id_p2n)
                if self.g.nodes[node]["bipartite"] == 1
                and self.g.nodes[node]["_obj"].label in location_labels
            )
        else:
            locations = (
                node
                for node in self.g.neighbors(actor.id_p2n)
                if self.g.nodes[node]["bipartite"] == 1
            )

        neighbor_actors = {
            actor_id
            for location_id in locations
            for actor_id in self.g.neighbors(location_id)
            if self.g.nodes[actor_id]["bipartite"] == 0
        }

        return self._to_framework(
            [
                self.g.nodes[actor_id]["_obj"]
                for actor_id in neighbor_actors
                if actor_id != actor.id_p2n
            ]
        )

    def _objects_between_objects(self, object1, object2) -> list:
        if object1 is object2:
            raise ValueError("Entity 1 and entity 2 are identical.")

        paths = list(
            nx.all_simple_paths(
                G=self.g,
                source=object1.id_p2n,
                target=object2.id_p2n,
                cutoff=2,
            ),
        )
        return [self.g.nodes[path[1]]["_obj"] for path in paths]

    def locations_between_actors(
        self, actor1, actor2, location_labels: list[str] | None = None
    ) -> list:
        """Return all locations that connects two actors.

        Args:
            actor1 (Actor): Actor 1.
            actor2 (Actor): Actor 2.
            location_labels (tuple, optional): Constrain the locations to the following types.
                Defaults to None.

        Returns:
            LocationList: A list of locations.
        """
        locations = self._objects_between_objects(object1=actor1, object2=actor2)

        if location_labels is not None:
            locations = [location for location in locations if location.label in location_labels]

        return self._to_framework(locations)

    def actors_between_locations(
        self, location1, location2, actor_types: list[str] | None = None
    ) -> list:
        """Return all actors between two locations.

        Args:
            location1 (Location): Location 1.
            location2 (Location): Location 2.
            actor_types (tuple, optional): Constrain the actors to the following types.
                Defaults to None.

        Returns:
            List: A list of actors.
        """
        actors = self._objects_between_objects(location1, location2)

        if actor_types is not None:
            actors = [actor for actor in actors if actor.type in actor_types]

        return self._to_framework(actors)

    def set_weight(self, actor, location, weight: float | None = None) -> None:
        """Set the weight of an actor at a location.

        If weight is None the method location.weight() will be used to generate a weight.

        Args:
            actor (Actor): The actor.
            location (Location): The location.
            weight (int): The weight
        """
        self.g[actor.id_p2n][location.id_p2n]["weight"] = (
            location.weight(actor) if weight is None else weight
        )

    def get_weight(self, actor, location) -> int:
        """Get the weight of an actor at a location.

        Args:
            actor (Actor): The actor.
            location (Location): The location.

        Returns:
            int: The weight.
        """
        return self.g[actor.id_p2n][location.id_p2n]["weight"]

    def connect_actors(self, actors: list, location_cls: type, weight: float | None = None):
        """Connects multiple actors via an instance of a given location class.

        Args:
            actors (list): A list of actors.
            location_cls (type): The location class that is used to create a location instance.
            weight (float | None): The edge weight between the actors and the location.
                Defaults to None.
        """
        if location_cls is None:
            if self.framework is None:
                location_cls = p2n.Location
            else:

                class Location(p2n.Location, self._framework.Agent):
                    pass

                location_cls = Location

        if self.model is None:
            location = location_cls()
        else:
            location = location_cls(model=self.model)

        self.add_location(location=location)
        location.add_actors(actors=actors, weight=weight)

    def disconnect_actors(
        self,
        actors: list,
        location_labels: list | None = None,
        remove_locations: bool = False,
    ):
        """Disconnects actors by removing them from shared locations.

        If a list of location types is given, only shared locations of the given types are
        considered. Turn on `remove_locations` in order to not only remove the given actors from the
        given location instance but also to remove the location instance from the environment.
        Use this method with care because removing actors from locations also disconnects those
        actors from all other actors connected to the location. Removing the location instance from
        the environment could have even more sideeffects to those actors still connected with this
        location!

        Args:
            actors (list): A list of actors.
            location_labels (list | None, optional): A list of location types to specify which
            shared locations are considered. Defaults to None.
            remove_locations (bool, optional): A bool that determines whether the shared locations
                shall be removed from the environment. Defaults to False.
        """
        pairs = list(itertools.combinations(actors, 2))

        shared_locations = []

        for actor1, actor2 in pairs:
            shared_locations.extend(
                self.locations_between_actors(
                    actor1=actor1,
                    actor2=actor2,
                    location_labels=location_labels,
                )
            )

        shared_locations = set(shared_locations)

        for location in shared_locations:
            warn = False
            for actor in location.actors:
                if actor not in actors:
                    warn = True
                    break

            if warn and self.enable_p2n_warnings:
                msg = "There are other actors at the location from which you have removed actors."
                warnings.warn(msg)

            location.remove_actors(actors)

            if remove_locations:
                self.remove_location(location=location)

                if warn and self.enable_p2n_warnings:
                    msg = "You have removed a location to which other actors were still connected."
                    warnings.warn(msg)

    def export_bipartite_network(
        self,
        actor_attrs: list | None = None,
        location_attrs: list | None = None,
    ) -> nx.graph:
        graph = self.g.copy()

        for i in graph:
            if graph.nodes[i]["bipartite"] == 0:
                if actor_attrs is not None:
                    for actor_attr in actor_attrs:
                        graph.nodes[i][actor_attr] = getattr(graph.nodes[i]["_obj"], actor_attr)

            elif graph.nodes[i]["bipartite"] == 1:
                if location_attrs is not None:
                    for location_attr in location_attrs:
                        graph.nodes[i][location_attr] = getattr(
                            graph.nodes[i]["_obj"], location_attr
                        )

            del graph.nodes[i]["_obj"]
        return graph

    def export_actor_network(
        self,
        node_attrs: list | None = None,
        include_0_weights: bool = True,
    ) -> nx.Graph:
        """Creates a projection of the environment's bipartite network.

        Args:
            node_attrs: A list of actor attributes
            include_0_weights: Should edges with weight 0 be displayed?

        Returns:
            A weighted graph created from a environment's actor list. Actors are connected if they are
            neighbors in the environment. Their connecting edge include the contact_weight as "weight"
            attribute.
        """
        graph = nx.Graph()

        # create nodes
        for actor in self.actors:
            if not graph.has_node(actor.id_p2n):
                node_attr_dict = (
                    {node_attr: vars(actor)[node_attr] for node_attr in node_attrs}
                    if node_attrs is not None
                    else {}
                )
                graph.add_node(actor.id_p2n, **node_attr_dict)

        # create edges
        for actor in self.actors:
            for actor_v in actor.neighbors():
                if not graph.has_edge(actor.id_p2n, actor_v.id_p2n):
                    weight = actor.get_actor_weight(actor_v)
                    if include_0_weights or weight > 0:
                        graph.add_edge(actor.id_p2n, actor_v.id_p2n, weight=weight)

        return graph

    def update_weights(self, location_labels: list | None = None) -> None:
        """Updates the edge weights between actors and locations.

        If you only want to update the weights of specific types of locations
        specify those types in location_labels.

        Args:
            location_labels (list | None, optional): A list of location classes that specifiy for
                which location types the weights should be updated.
                If location_labels is None all locations are considered. Defaults to None.
        """
        for location in (
            self.locations
            if location_labels is None
            else [location for location in self.locations if location.label in location_labels]
        ):
            for actor in location.actors:
                location.set_weight(actor=actor, weight=location.weight(actor=actor))



"""Helper to sample from a dataframe."""

from typing import Optional

import numpy as np
import pandas as pd


class DataReader:
    """Helper to sample from a dataframe."""

    def __init__(
        self,
        df: pd.DataFrame,
        seed: Optional[int] = None,
    ) -> None:
        """_summary_.

        Args:
            df (pd.DataFrame): _description_
            seed (Optional[int], optional): _description_. Defaults to None.
        """
        if seed:
            self.rng = np.random.default_rng(seed)
        else:
            self.rng = np.random.default_rng(None)

        self.df = df

    def sample(
        self,
        by: str,
        size: int,
        weights: Optional[str] = None,
        with_replacement: bool = True,
    ) -> pd.DataFrame:
        """_summary_.

        Args:
            by (str): _description_
            size (int): _description_
            weights (Optional[str], optional): _description_. Defaults to None.
            with_replacement (bool, optional): _description_. Defaults to True.

        Raises:
            ValueError: _description_

        Returns:
            pd.DataFrame: _description_
        """
        if weights is None:
            unique_by = self.df[[by]].drop_duplicates(subset=by, keep="first")
            weight_col = None
        else:
            unique_by = self.df[[by, weights]].drop_duplicates(subset=by, keep="first")
            weight_col = unique_by[weights]

        id_col = unique_by[by]
        if with_replacement or (size <= len(id_col)):
            sample = self.rng.choice(id_col, p=weight_col, replace=with_replacement, size=size)
        elif size <= len(self.df):
            # edge case:
            # sampling without replacement and non-unique sampling column while size
            # is larger than number of unique values to sample from
            sample = id_col.tolist()  # type: ignore
            self.rng.shuffle(sample)
        else:
            msg = f"Cannot sample size ({size}) larger than dataset ({len(self.df)}) without replacement!"  # noqa: E501
            raise ValueError(msg)

        if not with_replacement:
            # Faster shortcut that is possible if sampling is done without replacement.
            # Therefore implemented separately.
            df = self.df.loc[self.df[by].isin(sample), :]
            df = df.iloc[:size]
        else:
            subset_dfs = []
            len_counter = 0
            for sample_id in sample:
                subset = self.df[self.df[by] == sample_id]
                subset_dfs.append(subset)

                len_counter += len(subset)
                if len_counter >= size:
                    break

            df = pd.concat(subset_dfs, axis=0)
            df = df.iloc[:size]

        return df



from .actor import Actor
from .location import Location


class EntityList(list):
    """Simple list subclass for better print out."""

    def __str__(self):
        n = len(self)
        if n == 0:
            return "EntityList [ ]"

        actor_count = sum(isinstance(o, Actor) for o in self)
        location_count = sum(isinstance(o, Location) for o in self)
        unknown = n - actor_count - location_count

        parts = []
        if actor_count:
            parts.append(f"{actor_count} {'actor' if actor_count == 1 else 'actors'}")
        if location_count:
            parts.append(f"{location_count} {'location' if location_count == 1 else 'locations'}")
        if unknown:
            parts.append(f"{unknown} {'entity' if unknown == 1 else 'entities'}")

        return f"EntityList [{', '.join(parts)}]"

    def __repr__(self):
        return self.__str__()
